/**
 * \file	cf_file.c
 * \author	myusgun <myusgun@gmail.com>
 *
 * \brief 파일 입출력 구현
 */
#include "cf_file.h"
#include "cf_local.h"
#include "cf_error.h"

#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/stat.h>

#if defined(_WIN32) || defined(_WIN64)
# include <io.h>
# include <direct.h>

# define DELIMITER		'\\'
# define mkdir(a,b)		_mkdir (a)
# define access(a,b)	_access (a,b)

# define F_OK			0
# define W_OK			2
# define R_OK			4

/*------------------------------*/
# define S_IWUSR		_S_IWRITE
# define S_IRUSR		_S_IREAD
# define S_IXUSR		_S_IEXEC
/*------------------------------*/
# define S_IRGRP		0x00000000
# define S_IWGRP		0x00000000
# define S_IXGRP		0x00000000
/*------------------------------*/
# define S_IROTH		0x00000000
# define S_IWOTH		0x00000000
# define S_IXOTH		0x00000000
/*------------------------------*/
# define S_IRWXU		0x00000000
#else // #if defined(_WIN32) || defined(_WIN64)
# include <unistd.h>

# define DELIMITER		'/'
# define O_BINARY		0x00000000
#endif // #if defined(_WIN32) || defined(_WIN64)

#define DIRECTORY_MODE	S_IRWXU|S_IRGRP|S_IXGRP|S_IROTH|S_IXOTH
#define FILE_MODE		S_IWUSR|S_IRUSR|S_IRGRP|S_IROTH

#define ASSERT_ARGS(x)	\
	if ((x))			\
		return CF_ERROR_FILE_INVALID_ARGS

static int
CF_File_Local_ConvertFlag (const CF_FILE_FLAG flag)
{
	int result = 0;

	if (flag & CF_FILE_READ)	result |= O_RDONLY;
	if (flag & CF_FILE_WRITE)	result |= O_WRONLY;
	if (flag & CF_FILE_RW)		result |= O_RDWR;
	if (flag & CF_FILE_CREATE)	result |= O_CREAT;
	if (flag & CF_FILE_TRUNC)	result |= O_TRUNC;
	if (flag & CF_FILE_APPEND)	result |= O_APPEND;

	return result;
}

/**
 * 파일 열기
 *
 * \return 성공 시, 파일 디스크립터; 실패 시, 오류 코드
 *
 * \param path	파일 경로
 * \param flag	파일 열기 플래그
 *
 * \see CF_FILE_FLAG
 */
int
CF_File_Open (const char			* path,
			  const CF_FILE_FLAG	flag)
{
	int result = 0;
	int osflag = 0;

	ASSERT_ARGS (path == NULL);

	osflag = CF_File_Local_ConvertFlag (flag) | O_BINARY;

	result = open (path, osflag, FILE_MODE);
	if (result < 0)
		return CF_ERROR_FILE_OPEN;

	return result;
}

/**
 * 파일 생성
 *
 * \return 성공 시, 파일 디스크립터; 실패 시, 오류 코드
 *
 * \param path 파일 경로
 */
int
CF_File_Create (const char * path)
{
	int			result = 0;
	const int	flag = CF_FILE_CREATE|CF_FILE_WRITE|CF_FILE_TRUNC;
	
	ASSERT_ARGS (path == NULL);

	result = CF_File_Open (path, flag);
	if (result < 0)
		return CF_ERROR_FILE_CREATE;

	return result;
}

/**
 * 파일 닫기
 *
 * \return 성공 시, CF_OK; 실패 시, 오류 코드
 *
 * \param fd 파일 디스크립터
 */
int
CF_File_Close (const int fd)
{
	int result = 0;

	ASSERT_ARGS (fd < 0);

	result = close (fd);
	if (result < 0)
		return CF_ERROR_FILE_CLOSE;

	return CF_OK;
}

/**
 * 파일 읽기
 *
 * \return 성공 시, 읽은 바이트 수; 실패 시, 오류 코드
 *
 * \param fd	파일 디스크립터
 * \param buf	데이터를 저장할 메모리
 * \param len	데이터를 저장할 메모리의 크기
 */
int
CF_File_Read (const int		fd,
			  void			* buf,
			  const size_t	len)
{
	int result = 0;
	
	ASSERT_ARGS (fd < 0);
	ASSERT_ARGS (buf == NULL);

	result = (int) read (fd, buf, len);
	if (result < 0)
		return CF_ERROR_FILE_READ;

	return result;
}

/**
 * 파일 쓰기
 *
 * \return 성공 시, CF_OK; 실패 시, 오류 코드
 *
 * \param fd	파일 디스크립터
 * \param buf	데이터가 저장된 메모리
 * \param len	쓸 데이터의 길이
 */
int
CF_File_Write (const int	fd,
			   const void	* buf, 
			   const size_t	len)
{
	int result = 0;
	
	ASSERT_ARGS (fd < 0);
	ASSERT_ARGS (buf == NULL);

	result = (int) write (fd, buf, len);
	if (result != len)
	  return CF_ERROR_FILE_WRITE;

	return CF_OK;
}

/**
 * 파일 크기 얻기
 *
 * \return 성공 시, 파일 크기; 실패 시, 오류 코드
 *
 * \param path 파일 경로
 */
int
CF_File_GetSize (const char * path)
{
	int length = 0;
	int fd = 0;

	ASSERT_ARGS (path == NULL);

	fd = CF_File_Open (path, CF_FILE_READ);
	if (fd < 0)
		return CF_ERROR_FILE_OPEN;

	length = (int) lseek (fd, 0, SEEK_END);
	if (length < 0)
		return CF_ERROR_FILE_GET_SIZE;

	CF_File_Close (fd);

	return length;
}

/**
 * 파일 및 디렉터리 존재 여부 검사
 *
 * \return 존재 시, CF_TRUE; 아니면, CF_FALSE
 *
 * \param path 파일 및 디렉터리 경로
 */
CF_BOOL
CF_File_Exists (const char * path)
{
	return (access (path, F_OK) == 0) ? CF_TRUE : CF_FALSE;
}

/**
 * 디렉터리 생성
 *
 * \return 성공 시, CF_OK; 실패 시, 오류 코드
 *
 * \param path 생성할 디렉터리 경로
 */
int
CF_File_MakeDirectory (const char * path)
{
	int		result = 0;
	char	fullPath[1024] = {0x00,};
	char	stepPath[256] = {0x00,};

	char	* f = fullPath;
	char	* d = stepPath;

	ASSERT_ARGS (path == NULL);

	snprintf (fullPath, sizeof (fullPath) - 1, "%s%c", path, DELIMITER);

	for (*d++ = *f++ ; *f ; *d++ = *f++)
	{
		if (*f != DELIMITER)
			continue;

		if (CF_File_Exists (stepPath))
			continue;

		result = mkdir (stepPath, DIRECTORY_MODE);
		if (result && errno != EEXIST)
			return CF_ERROR_FILE_MAKE_DIRECTORY;
	}

	return CF_OK;
}
